import numpy as np
from pandas.testing import assert_frame_equal

import pyranges as pr

ensembl_gtf = "tests/unit/test_data/ensembl.gtf"


def test_read_gtf():
    gr = pr.read_gtf(ensembl_gtf, full=True)

    assert len(gr.columns) == 26

    df = gr.df
    transcript = df.iloc[1]
    assert transcript["tag"] == "basic"

    exon = df[df["exon_id"] == "ENSE00003812156"].iloc[0]
    assert exon["tag"] == "basic"

    gr = pr.read_gtf(ensembl_gtf, full=True, duplicate_attr=True)
    print(gr.columns)
    assert len(gr.columns) == 26

    df = gr.df
    transcript = df.iloc[1]
    assert transcript["tag"] == "basic"

    exon = df[df["exon_id"] == "ENSE00003812156"].iloc[0]
    assert exon["tag"] == "CCDS,basic"
    # assert list(gr.df.columns[:4]) == "Chromosome Start End Strand".split()


def test_read_gtf_multiple_chunks():
    skiprows = pr.readers.skiprows(ensembl_gtf)
    gr = pr.readers.read_gtf_full(ensembl_gtf, chunksize=2, skiprows=skiprows)
    gr2 = pr.readers.read_gtf_full(ensembl_gtf, skiprows=skiprows)

    df = gr.df
    df2 = gr2.df
    # reading in chunks might mix types because reading a chunk of . 0 will make each value a str
    # however, when reading a chunk of 0 0 the type will be int
    df.loc[:, "Frame"] = df["Frame"].astype(str)

    df.Feature = np.array(df["Feature"].astype(str))
    df2.Feature = np.array(df2["Feature"].astype(str))

    assert_frame_equal(df, df2, check_dtype=False)


def test_read_gff3():
    gr = pr.read_gff3("tests/unit/test_data/gencode.gff3")

    assert len(gr.columns) == 26
    # assert list(gr.df.columns[:4]) == "Chromosome Start End Strand".split()


def test_read_bed():
    pr.read_bed("pyranges/example_data/chipseq.bed")



import numpy as np
import pandas as pd
from sorted_nearest.src.introns import find_introns  # type: ignore

import pyranges as pr
from pyranges.multithreaded import pyrange_apply

__all__ = ["genome_bounds", "tile_genome", "GenomicFeaturesMethods"]


class GenomicFeaturesMethods:

    """Namespace for methods using feature information.

    Accessed through `gr.features`."""

    pr = None

    def __init__(self, pr):
        self.pr = pr

    def tss(self):
        """Return the transcription start sites.

        Returns the 5' for every interval with feature "transcript".

        See Also
        --------
        pyranges.genomicfeatures.GenomicFeaturesMethods.tes : return the transcription end sites

        Examples
        --------

        >>> gr = pr.data.ensembl_gtf()[["Source", "Feature"]]
        >>> gr
        +--------------+------------+--------------+-----------+-----------+--------------+
        | Chromosome   | Source     | Feature      | Start     | End       | Strand       |
        | (category)   | (object)   | (category)   | (int64)   | (int64)   | (category)   |
        |--------------+------------+--------------+-----------+-----------+--------------|
        | 1            | havana     | gene         | 11868     | 14409     | +            |
        | 1            | havana     | transcript   | 11868     | 14409     | +            |
        | 1            | havana     | exon         | 11868     | 12227     | +            |
        | 1            | havana     | exon         | 12612     | 12721     | +            |
        | ...          | ...        | ...          | ...       | ...       | ...          |
        | 1            | havana     | gene         | 1173055   | 1179555   | -            |
        | 1            | havana     | transcript   | 1173055   | 1179555   | -            |
        | 1            | havana     | exon         | 1179364   | 1179555   | -            |
        | 1            | havana     | exon         | 1173055   | 1176396   | -            |
        +--------------+------------+--------------+-----------+-----------+--------------+
        Stranded PyRanges object has 2,446 rows and 6 columns from 1 chromosomes.
        For printing, the PyRanges was sorted on Chromosome and Strand.

        >>> gr.features.tss()
        +--------------+------------+------------+-----------+-----------+--------------+
        | Chromosome   | Source     | Feature    | Start     | End       | Strand       |
        | (category)   | (object)   | (object)   | (int64)   | (int64)   | (category)   |
        |--------------+------------+------------+-----------+-----------+--------------|
        | 1            | havana     | tss        | 11868     | 11869     | +            |
        | 1            | havana     | tss        | 12009     | 12010     | +            |
        | 1            | havana     | tss        | 29553     | 29554     | +            |
        | 1            | havana     | tss        | 30266     | 30267     | +            |
        | ...          | ...        | ...        | ...       | ...       | ...          |
        | 1            | havana     | tss        | 1092812   | 1092813   | -            |
        | 1            | havana     | tss        | 1116086   | 1116087   | -            |
        | 1            | havana     | tss        | 1116088   | 1116089   | -            |
        | 1            | havana     | tss        | 1179554   | 1179555   | -            |
        +--------------+------------+------------+-----------+-----------+--------------+
        Stranded PyRanges object has 280 rows and 6 columns from 1 chromosomes.
        For printing, the PyRanges was sorted on Chromosome and Strand.
        """

        pr = self.pr

        if not pr.stranded:
            raise Exception(
                "Cannot compute TSSes or TESes without strand info. Perhaps use extend() or subsequence() or spliced_subsequence() instead?"
            )

        pr = pr[pr.Feature == "transcript"]
        pr = pr.apply(lambda df: _tss(df))

        pr.Feature = "tss"

        return pr

    def tes(self, slack=0):
        """Return the transcription end sites.

        Returns the 3' for every interval with feature "transcript".

        See Also
        --------
        pyranges.genomicfeatures.GenomicFeaturesMethods.tss : return the transcription start sites

        Examples
        --------

        >>> gr = pr.data.ensembl_gtf()[["Source", "Feature"]]
        >>> gr
        +--------------+------------+--------------+-----------+-----------+--------------+
        | Chromosome   | Source     | Feature      | Start     | End       | Strand       |
        | (category)   | (object)   | (category)   | (int64)   | (int64)   | (category)   |
        |--------------+------------+--------------+-----------+-----------+--------------|
        | 1            | havana     | gene         | 11868     | 14409     | +            |
        | 1            | havana     | transcript   | 11868     | 14409     | +            |
        | 1            | havana     | exon         | 11868     | 12227     | +            |
        | 1            | havana     | exon         | 12612     | 12721     | +            |
        | ...          | ...        | ...          | ...       | ...       | ...          |
        | 1            | havana     | gene         | 1173055   | 1179555   | -            |
        | 1            | havana     | transcript   | 1173055   | 1179555   | -            |
        | 1            | havana     | exon         | 1179364   | 1179555   | -            |
        | 1            | havana     | exon         | 1173055   | 1176396   | -            |
        +--------------+------------+--------------+-----------+-----------+--------------+
        Stranded PyRanges object has 2,446 rows and 6 columns from 1 chromosomes.
        For printing, the PyRanges was sorted on Chromosome and Strand.

        >>> gr.features.tes()
        +--------------+------------+------------+-----------+-----------+--------------+
        | Chromosome   | Source     | Feature    | Start     | End       | Strand       |
        | (category)   | (object)   | (object)   | (int64)   | (int64)   | (category)   |
        |--------------+------------+------------+-----------+-----------+--------------|
        | 1            | havana     | tes        | 14408     | 14409     | +            |
        | 1            | havana     | tes        | 13669     | 13670     | +            |
        | 1            | havana     | tes        | 31096     | 31097     | +            |
        | 1            | havana     | tes        | 31108     | 31109     | +            |
        | ...          | ...        | ...        | ...       | ...       | ...          |
        | 1            | havana     | tes        | 1090405   | 1090406   | -            |
        | 1            | havana     | tes        | 1091045   | 1091046   | -            |
        | 1            | havana     | tes        | 1091499   | 1091500   | -            |
        | 1            | havana     | tes        | 1173055   | 1173056   | -            |
        +--------------+------------+------------+-----------+-----------+--------------+
        Stranded PyRanges object has 280 rows and 6 columns from 1 chromosomes.
        For printing, the PyRanges was sorted on Chromosome and Strand.
        """

        pr = self.pr

        if not pr.stranded:
            raise Exception(
                "Cannot compute TSSes or TESes without strand info. Perhaps use extend() or subsequence() or spliced_subsequence() instead?"
            )

        pr = pr[pr.Feature == "transcript"]
        pr = pr.apply(lambda df: _tes(df))

        pr.Feature = "tes"

        return pr

    def introns(self, by="gene", nb_cpu=1):
        """Return the introns.

        Parameters
        ----------
        by : str, {"gene", "transcript"}, default "gene"
            Whether to find introns per gene or transcript.

        nb_cpu: int, default 1

            How many cpus to use. Can at most use 1 per chromosome or chromosome/strand tuple.
            Will only lead to speedups on large datasets.

        See Also
        --------
        pyranges.genomicfeatures.GenomicFeaturesMethods.tss : return the transcription start sites

        Examples
        --------

        >>> gr = pr.data.ensembl_gtf()[["Feature", "gene_id", "transcript_id"]]
        >>> gr
        +--------------+--------------+-----------+-----------+--------------+-----------------+-----------------+
        | Chromosome   | Feature      | Start     | End       | Strand       | gene_id         | transcript_id   |
        | (category)   | (category)   | (int64)   | (int64)   | (category)   | (object)        | (object)        |
        |--------------+--------------+-----------+-----------+--------------+-----------------+-----------------|
        | 1            | gene         | 11868     | 14409     | +            | ENSG00000223972 | nan             |
        | 1            | transcript   | 11868     | 14409     | +            | ENSG00000223972 | ENST00000456328 |
        | 1            | exon         | 11868     | 12227     | +            | ENSG00000223972 | ENST00000456328 |
        | 1            | exon         | 12612     | 12721     | +            | ENSG00000223972 | ENST00000456328 |
        | ...          | ...          | ...       | ...       | ...          | ...             | ...             |
        | 1            | gene         | 1173055   | 1179555   | -            | ENSG00000205231 | nan             |
        | 1            | transcript   | 1173055   | 1179555   | -            | ENSG00000205231 | ENST00000379317 |
        | 1            | exon         | 1179364   | 1179555   | -            | ENSG00000205231 | ENST00000379317 |
        | 1            | exon         | 1173055   | 1176396   | -            | ENSG00000205231 | ENST00000379317 |
        +--------------+--------------+-----------+-----------+--------------+-----------------+-----------------+
        Stranded PyRanges object has 2,446 rows and 7 columns from 1 chromosomes.
        For printing, the PyRanges was sorted on Chromosome and Strand.

        >>> gr.features.introns(by="gene")
        +--------------+------------+-----------+-----------+--------------+-----------------+-----------------+
        | Chromosome   | Feature    | Start     | End       | Strand       | gene_id         | transcript_id   |
        | (object)     | (object)   | (int64)   | (int64)   | (category)   | (object)        | (object)        |
        |--------------+------------+-----------+-----------+--------------+-----------------+-----------------|
        | 1            | intron     | 1173926   | 1174265   | +            | ENSG00000162571 | nan             |
        | 1            | intron     | 1174321   | 1174423   | +            | ENSG00000162571 | nan             |
        | 1            | intron     | 1174489   | 1174520   | +            | ENSG00000162571 | nan             |
        | 1            | intron     | 1175034   | 1179188   | +            | ENSG00000162571 | nan             |
        | ...          | ...        | ...       | ...       | ...          | ...             | ...             |
        | 1            | intron     | 874591    | 875046    | -            | ENSG00000283040 | nan             |
        | 1            | intron     | 875155    | 875525    | -            | ENSG00000283040 | nan             |
        | 1            | intron     | 875625    | 876526    | -            | ENSG00000283040 | nan             |
        | 1            | intron     | 876611    | 876754    | -            | ENSG00000283040 | nan             |
        +--------------+------------+-----------+-----------+--------------+-----------------+-----------------+
        Stranded PyRanges object has 311 rows and 7 columns from 1 chromosomes.
        For printing, the PyRanges was sorted on Chromosome and Strand.

        >>> gr.features.introns(by="transcript")
        +--------------+------------+-----------+-----------+--------------+-----------------+-----------------+
        | Chromosome   | Feature    | Start     | End       | Strand       | gene_id         | transcript_id   |
        | (object)     | (object)   | (int64)   | (int64)   | (category)   | (object)        | (object)        |
        |--------------+------------+-----------+-----------+--------------+-----------------+-----------------|
        | 1            | intron     | 818202    | 818722    | +            | ENSG00000177757 | ENST00000326734 |
        | 1            | intron     | 960800    | 961292    | +            | ENSG00000187961 | ENST00000338591 |
        | 1            | intron     | 961552    | 961628    | +            | ENSG00000187961 | ENST00000338591 |
        | 1            | intron     | 961750    | 961825    | +            | ENSG00000187961 | ENST00000338591 |
        | ...          | ...        | ...       | ...       | ...          | ...             | ...             |
        | 1            | intron     | 732207    | 732980    | -            | ENSG00000230021 | ENST00000648019 |
        | 1            | intron     | 168165    | 169048    | -            | ENSG00000241860 | ENST00000655252 |
        | 1            | intron     | 165942    | 167958    | -            | ENSG00000241860 | ENST00000662089 |
        | 1            | intron     | 168165    | 169048    | -            | ENSG00000241860 | ENST00000662089 |
        +--------------+------------+-----------+-----------+--------------+-----------------+-----------------+
        Stranded PyRanges object has 1,043 rows and 7 columns from 1 chromosomes.
        For printing, the PyRanges was sorted on Chromosome and Strand.
        """

        kwargs = {"by": by, "nb_cpu": nb_cpu}
        kwargs = pr.pyranges_main.fill_kwargs(kwargs)

        assert by in ["gene", "transcript"]

        id_column = by_to_id[by]
        gr = self.pr.sort(id_column)

        if not len(gr):
            return pr.PyRanges()

        exons = gr.subset(lambda df: df.Feature == "exon")
        exons = exons.merge(by=id_column)

        by_gr = gr.subset(lambda df: df.Feature == by)

        result = pyrange_apply(_introns2, by_gr, exons, **kwargs)

        return pr.PyRanges(result)


def _outside_bounds(df, **kwargs):
    df = df.copy()

    chromsizes = kwargs.get("chromsizes")

    if not isinstance(chromsizes, dict):
        size_df = chromsizes.df
        chromsizes = {k: v for k, v in zip(size_df.Chromosome, size_df.End)}

    size = int(chromsizes[df.Chromosome.iloc[0]])
    clip = kwargs.get("clip", False)
    only_right = kwargs.get("only_right", False)

    ends_outright = df.End > size
    if not only_right:
        starts_outleft = df.Start < 0

    if not clip:  # i.e. remove
        if only_right:
            df = df[~ends_outright]
        else:
            df = df[~ends_outright & ~starts_outleft]

    else:
        starts_outright = df.Start >= size

        if only_right:
            df.loc[ends_outright, "End"] = size

            # removing intervals completely out of bounds
            df = df[~starts_outright]

        else:
            ends_outleft = df.End <= 0

            df.loc[ends_outright, "End"] = size
            df.loc[starts_outleft, "Start"] = 0

            # removing intervals completely out of bounds:
            df = df[~starts_outright & ~ends_outleft]

    return df


def genome_bounds(gr, chromsizes, clip=False, only_right=False):
    """Remove or clip intervals outside of genome bounds.

    Parameters
    ----------

    gr : PyRanges

        Input intervals

    chromsizes : dict or PyRanges or pyfaidx.Fasta

        Dict or PyRanges describing the lengths of the chromosomes.
        pyfaidx.Fasta object is also accepted since it conveniently loads chromosome length

    clip : bool, default False

        Returns the portions of intervals within bounds,
        instead of dropping intervals entirely if they are even partially
        out of bounds

    only_right : bool, default False

        If True, remove or clip only intervals that are out-of-bounds on the right,
        and do not alter those out-of-bounds on the left (whose Start is < 0)


    Examples
    --------

    >>> d = {"Chromosome": [1, 1, 3], "Start": [1, 249250600, 5], "End": [2, 249250640, 7]}
    >>> gr = pr.from_dict(d)
    >>> gr
    +--------------+-----------+-----------+
    |   Chromosome |     Start |       End |
    |   (category) |   (int64) |   (int64) |
    |--------------+-----------+-----------|
    |            1 |         1 |         2 |
    |            1 | 249250600 | 249250640 |
    |            3 |         5 |         7 |
    +--------------+-----------+-----------+
    Unstranded PyRanges object has 3 rows and 3 columns from 2 chromosomes.
    For printing, the PyRanges was sorted on Chromosome.

    >>> chromsizes = {"1": 249250621, "3": 500}
    >>> chromsizes
    {'1': 249250621, '3': 500}

    >>> pr.gf.genome_bounds(gr, chromsizes)
    +--------------+-----------+-----------+
    |   Chromosome |     Start |       End |
    |   (category) |   (int64) |   (int64) |
    |--------------+-----------+-----------|
    |            1 |         1 |         2 |
    |            3 |         5 |         7 |
    +--------------+-----------+-----------+
    Unstranded PyRanges object has 2 rows and 3 columns from 2 chromosomes.
    For printing, the PyRanges was sorted on Chromosome.

    >>> pr.gf.genome_bounds(gr, chromsizes, clip=True)
    +--------------+-----------+-----------+
    |   Chromosome |     Start |       End |
    |   (category) |   (int64) |   (int64) |
    |--------------+-----------+-----------|
    |            1 |         1 |         2 |
    |            1 | 249250600 | 249250621 |
    |            3 |         5 |         7 |
    +--------------+-----------+-----------+
    Unstranded PyRanges object has 3 rows and 3 columns from 2 chromosomes.
    For printing, the PyRanges was sorted on Chromosome.

    >>> del chromsizes['3']
    >>> chromsizes
    {'1': 249250621}

    >>> pr.gf.genome_bounds(gr, chromsizes)
    Traceback (most recent call last):
    ...
    KeyError: '3'
    """

    if isinstance(chromsizes, pr.PyRanges):
        chromsizes = {k: v for k, v in zip(chromsizes.Chromosome, chromsizes.End)}

    elif isinstance(chromsizes, dict):
        pass

    else:
        try:
            import pyfaidx  # type: ignore

            if isinstance(chromsizes, pyfaidx.Fasta):
                chromsizes = {k: len(chromsizes[k]) for k in chromsizes.keys()}
        except ImportError:
            pass

    assert isinstance(
        chromsizes, dict
    ), "ERROR chromsizes must be a dictionary, or a PyRanges, or a pyfaidx.Fasta object"

    return gr.apply(_outside_bounds, chromsizes=chromsizes, clip=clip, only_right=only_right)


def _last_tile(df, **kwargs):
    # do not need copy, since it is only used internally by
    # tile_genome
    # df = df.copy()
    sizes = kwargs.get("sizes")
    size = sizes[df.Chromosome.iloc[0]].End.iloc[0]
    df.loc[df.tail(1).index, "End"] = size

    return df


def tile_genome(genome, tile_size, tile_last=False):
    """Create a tiled genome.

    Parameters
    ----------
    chromsizes : dict or PyRanges

        Dict or PyRanges describing the lengths of the chromosomes.

    tile_size : int
        Length of the tiles.

    tile_last : bool, default False

        Use genome length as end of last tile.

    See Also
    --------

    pyranges.PyRanges.tile : split intervals into adjacent non-overlapping tiles.

    Examples
    --------

    >>> chromsizes = pr.data.chromsizes()
    >>> chromsizes
    +--------------+-----------+-----------+
    | Chromosome   | Start     | End       |
    | (category)   | (int64)   | (int64)   |
    |--------------+-----------+-----------|
    | chr1         | 0         | 249250621 |
    | chr2         | 0         | 243199373 |
    | chr3         | 0         | 198022430 |
    | chr4         | 0         | 191154276 |
    | ...          | ...       | ...       |
    | chr22        | 0         | 51304566  |
    | chrM         | 0         | 16571     |
    | chrX         | 0         | 155270560 |
    | chrY         | 0         | 59373566  |
    +--------------+-----------+-----------+
    Unstranded PyRanges object has 25 rows and 3 columns from 25 chromosomes.
    For printing, the PyRanges was sorted on Chromosome.

    >>> pr.gf.tile_genome(chromsizes, int(1e6))
    +--------------+-----------+-----------+
    | Chromosome   | Start     | End       |
    | (category)   | (int64)   | (int64)   |
    |--------------+-----------+-----------|
    | chr1         | 0         | 1000000   |
    | chr1         | 1000000   | 2000000   |
    | chr1         | 2000000   | 3000000   |
    | chr1         | 3000000   | 4000000   |
    | ...          | ...       | ...       |
    | chrY         | 56000000  | 57000000  |
    | chrY         | 57000000  | 58000000  |
    | chrY         | 58000000  | 59000000  |
    | chrY         | 59000000  | 59373566  |
    +--------------+-----------+-----------+
    Unstranded PyRanges object has 3,114 rows and 3 columns from 25 chromosomes.
    For printing, the PyRanges was sorted on Chromosome.
    """

    if isinstance(genome, dict):
        chromosomes, ends = list(genome.keys()), list(genome.values())
        df = pd.DataFrame({"Chromosome": chromosomes, "Start": 0, "End": ends})
        genome = pr.PyRanges(df)

    gr = genome.tile(tile_size)

    if not tile_last:
        gr = gr.apply(_last_tile, sizes=genome)

    return gr


def _keep_transcript_with_most_exons(df):
    transcripts_with_most_exons = []

    for _, gdf in df.groupby("gene_id"):
        max_exon = gdf.exon_number.max()
        max_transcript = gdf.loc[gdf.exon_number == max_exon].Transcript.iloc[0]

        max_rows = gdf.loc[gdf.Transcript == max_transcript]

        transcripts_with_most_exons.append(max_rows)

    return pd.concat(transcripts_with_most_exons).reset_index(drop=True)


def filter_transcripts(df, keep="most_exons"):
    return _keep_transcript_with_most_exons(df)


def _tss(df, slack=0):
    intype = df.Start.dtype

    tss_pos = df.loc[df.Strand == "+"]

    tss_neg = df.loc[df.Strand == "-"].copy()

    # pd.options.mode.chained_assignment = None
    tss_neg.loc[:, "Start"] = tss_neg.End - 1

    # pd.options.mode.chained_assignment = "warn"
    tss = pd.concat([tss_pos, tss_neg], sort=False)
    tss["End"] = tss.Start + 1
    tss.End = tss.End + slack
    tss.Start = tss.Start - slack
    tss.loc[tss.Start < 0, "Start"] = 0

    tss.index = range(len(tss))

    tss[["Start", "End"]] = tss[["Start", "End"]].astype(intype)

    return tss


def _tes(df, slack=0):
    intype = df.Start.dtype
    # df = self.df

    tes_pos = df.loc[df.Strand == "+"]

    tes_neg = df.loc[df.Strand == "-"].copy()

    # pd.options.mode.chained_assignment = None
    tes_neg.loc[:, "End"] = tes_neg.Start + 1

    # pd.options.mode.chained_assignment = "warn"
    tes = pd.concat([tes_pos, tes_neg], sort=False)
    tes["Start"] = tes.End - 1
    tes.End = tes.End + slack
    tes.Start = tes.Start - slack
    tes.loc[tes.Start < 0, "Start"] = 0

    tes.index = range(len(tes))

    tes[["Start", "End"]] = tes[["Start", "End"]].astype(intype)

    return tes


by_to_id = {"gene": "gene_id", "transcript": "transcript_id"}


def _introns2(df, exons, **kwargs):
    """TODO: refactor"""

    if df.empty or exons.empty:
        return None

    original_order = df.columns
    by = kwargs["by"]
    id_column = by_to_id[by]

    exons = exons[["Start", "End", id_column]]
    genes = df[["Start", "End", id_column]]
    exons.columns = ["Start", "End", "by_id"]
    genes.columns = ["Start", "End", "by_id"]

    intersection = pd.Series(np.intersect1d(exons["by_id"], genes["by_id"]))
    if len(intersection) == 0:
        return None

    exons = exons[exons["by_id"].isin(intersection)].reset_index(drop=True).sort_values(["by_id", "Start"])
    genes = genes[genes["by_id"].isin(intersection)].reset_index(drop=True).sort_values(["by_id", "Start"])
    df = df[df[id_column].isin(intersection)].reset_index(drop=True)

    assert len(genes) == len(
        genes.drop_duplicates("by_id")
    ), "The {id_column}s need to be unique to compute the introns.".format(id_column=id_column)

    exon_ids = exons["by_id"].shift() != exons["by_id"]
    by_ids = pd.Series(range(1, len(genes) + 1))
    df.insert(0, "__temp__", by_ids)

    if len(exons) > 1 and exons["by_id"].iloc[0] == exons["by_id"].iloc[1]:
        exon_ids.iloc[0] = False
        exon_ids = exon_ids.cumsum() + 1
    else:
        exon_ids = exon_ids.cumsum()

    assert (by_ids == exon_ids.drop_duplicates().values).all()
    starts, ends, ids = find_introns(
        genes.Start.values,
        genes.End.values,
        by_ids.values,
        exons.Start.values,
        exons.End.values,
        exon_ids.values,
    )

    introns = pd.DataFrame(
        data={
            "Chromosome": df.Chromosome.iloc[0],
            "Start": starts,
            "End": ends,
            "by_id": ids,
        }
    )

    vc = introns["by_id"].value_counts(sort=False).to_frame().reset_index()
    vc.columns = ["by_id", "counts"]

    genes_without_introns = pd.DataFrame(data={"by_id": np.setdiff1d(by_ids.values, vc.by_id.values), "counts": 0})

    vc = pd.concat([vc, genes_without_introns]).sort_values("by_id")

    original_ids = np.repeat(vc.by_id, vc.counts).to_frame()
    original_ids = original_ids.merge(
        df[["__temp__", id_column]],
        right_on="__temp__",
        left_on="by_id",
        suffixes=("_drop", ""),
    )
    original_ids = original_ids.drop(
        ["__temp__"] + [c for c in original_ids.columns if c.endswith("_drop")], axis=1
    ).sort_values("by_id")
    introns.loc[:, "by_id"] = original_ids[id_column].values
    introns = introns.merge(df, left_on="by_id", right_on=id_column, suffixes=("", "_dropme"))
    introns = introns.drop([c for c in introns.columns if c.endswith("_dropme")], axis=1)

    if introns.Feature.dtype.name == "category" and "intron" not in introns.Feature.cat.categories:
        introns.Feature.cat.add_categories(["intron"])
    introns.loc[:, "Feature"] = "intron"

    introns = introns[original_order]

    return introns



import yaml

rmds = [
    "index.Rmd",
    "creating.Rmd",
    "writing.Rmd",
    "subset.Rmd",
    "manipulation.Rmd",
    "concat.Rmd",
    "piping.Rmd",
    "iterate.Rmd",
    "sort.Rmd",
    "summary.Rmd",
    "single_range_methods.Rmd",
    "intersection.Rmd",
    "overlap.Rmd",
    "join.Rmd",
    "nearest.Rmd",
    "jaccard.Rmd",
    "reldist.Rmd",
    "coverage.Rmd",
    "runlengths.Rmd",
    "runlength_dict.Rmd",
    "subsetting_rles.Rmd",
    "subsetting_pyrles.Rmd",
    "multithreading.Rmd",
    "databases.Rmd",
]

from subprocess import call

single_page = "\n".join(open(f).read() for f in rmds[1:])
import re

single_page = re.sub(re.compile("#"), "##", single_page)
single_page = open(rmds[0]).read() + "\n" + single_page
open("one_page.Rmd", "w+").write(single_page)

# cmd = "cat {} > one_page.Rmd".format(" ".join(rmds))
# call("cat {} > one_page.Rmd".format(" ".join(rmds)), shell=True)

call("Rscript compile2.R", shell=True)



import numpy as np
import pandas as pd

from pyranges import PyRanges
from pyranges.genomicfeatures import GenomicFeaturesMethods
from pyranges.helpers import get_key_from_df, single_value_key
from pyranges.statistics import StatisticsMethods


def set_dtypes(df):
    from pandas.api.types import CategoricalDtype

    dtypes = {
        "Start": np.int64,
        "End": np.int64,
        "Chromosome": "category",
    }

    if "Strand" in df:
        dtypes["Strand"] = CategoricalDtype(categories=df["Strand"].drop_duplicates().to_list())

    # need to ascertain that object columns do not consist of multiple types
    # https://github.com/biocore-ntnu/epic2/issues/32
    for column in "Chromosome Strand".split():
        if column not in df:
            continue
        df[column] = df[column].astype(str)

    for col, dtype in dtypes.items():
        if df[col].dtype.name != dtype:
            df[col] = df[col].astype(dtype)
    return df


def create_df_dict(df, stranded):
    chrs = df.Chromosome.cat.remove_unused_categories()

    df["Chromosome"] = chrs

    if stranded:
        grpby_key = "Chromosome Strand".split()
    else:
        grpby_key = "Chromosome"

    return {k: v for k, v in df.groupby(grpby_key)}


def create_pyranges_df(chromosomes, starts, ends, strands=None):
    if isinstance(chromosomes, str) or isinstance(chromosomes, int):
        chromosomes = pd.Series([chromosomes] * len(starts), dtype="category")

    if strands is not None:
        if isinstance(strands, str):
            strands = pd.Series([strands] * len(starts), dtype="category")

        columns = [chromosomes, starts, ends, strands]
        lengths = list(str(len(s)) for s in columns)
        assert (
            len(set(lengths)) == 1
        ), "chromosomes, starts, ends and strands must be of equal length. But are {}".format(", ".join(lengths))
        colnames = "Chromosome Start End Strand".split()
    else:
        columns = [chromosomes, starts, ends]
        lengths = list(str(len(s)) for s in columns)
        assert len(set(lengths)) == 1, "chromosomes, starts and ends must be of equal length. But are {}".format(
            ", ".join(lengths)
        )
        colnames = "Chromosome Start End".split()

    idx = range(len(starts))
    series_to_concat = []
    for s in columns:
        if isinstance(s, pd.Series):
            s = pd.Series(s.values, index=idx)
        else:
            s = pd.Series(s, index=idx)

        series_to_concat.append(s)

    df = pd.concat(series_to_concat, axis=1)
    df.columns = colnames

    return df


def check_strandedness(df):
    """Check whether strand contains '.'"""

    if "Strand" not in df:
        return False

    contains_more_than_plus_minus_in_strand_col = False

    if str(df.Strand.dtype) == "category" and (set(df.Strand.cat.categories) - set("+-")):
        contains_more_than_plus_minus_in_strand_col = True
    elif not ((df.Strand == "+") | (df.Strand == "-")).all():
        contains_more_than_plus_minus_in_strand_col = True

        # if contains_more_than_plus_minus_in_strand_col:
        #     logging.warning("Strand contained more symbols than '+' or '-'. Not supported (yet) in PyRanges.")

    return not contains_more_than_plus_minus_in_strand_col


def _init(
    self,
    df=None,
    chromosomes=None,
    starts=None,
    ends=None,
    strands=None,
    copy_df=True,
):
    # TODO: add categorize argument with dict of args to categorize?

    if isinstance(df, PyRanges):
        raise Exception("Object is already a PyRange.")

    if isinstance(df, pd.DataFrame):
        assert all(
            c in df for c in "Chromosome Start End".split()
        ), "The dataframe does not have all the columns Chromosome, Start and End."
        if copy_df:
            df = df.copy()

    if df is False or df is None:
        df = create_pyranges_df(chromosomes, starts, ends, strands)

    if isinstance(df, pd.DataFrame):
        df = df.reset_index(drop=True)

        stranded = check_strandedness(df)

        df = set_dtypes(df)

        self.__dict__["dfs"] = create_df_dict(df, stranded)

    # df is actually dict of dfs
    else:
        empty_removed = {k: v.copy() for k, v in df.items() if not v.empty}

        _single_value_key = True
        _key_same = True
        _strand_valid = True
        _has_strand = True
        for key, df in empty_removed.items():
            _key = get_key_from_df(df)
            _single_value_key = single_value_key(df) and _single_value_key
            _key_same = (_key == key) and _key_same

            if isinstance(_key, tuple):
                _strand_valid = _strand_valid and (_key[1] in ["+", "-"])
            else:
                _has_strand = False

        if not all([_single_value_key, _key_same, _strand_valid]):
            df = pd.concat(empty_removed.values()).reset_index(drop=True)

            if _has_strand and _strand_valid:
                empty_removed = df.groupby(["Chromosome", "Strand"])
            else:
                empty_removed = df.groupby("Chromosome")

            empty_removed = {k: v for (k, v) in empty_removed}

        self.__dict__["dfs"] = empty_removed

    self.__dict__["features"] = GenomicFeaturesMethods(self)
    self.__dict__["stats"] = StatisticsMethods(self)



import pandas as pd

import pyranges as pr


def test_change_chromosomes():
    df1 = pd.DataFrame({"Chromosome": ["chr1", "chr2"], "Start": [100, 200], "End": [150, 201]})
    py1 = pr.PyRanges(df1)

    def modify_chrom_series(df):
        df.Chromosome = df.Chromosome.apply(lambda val: val.replace("chr", ""))
        return df

    def fix_chrom(regions):
        return regions.apply(modify_chrom_series)

    print(py1)

    py1 = fix_chrom(py1)

    assert py1.chromosomes == ["1", "2"]



import pickle

from pandas.testing import assert_frame_equal

import pyranges as pr


def test_pickle():
    gr = pr.data.f1()
    pickle.dump(gr, open("hi", "wb+"))
    gr2 = pickle.load(open("hi", "rb"))
    assert_frame_equal(gr.df, gr2.df)



import csv

import numpy as np
import pandas as pd
from natsort import natsorted  # type: ignore

_gtf_columns = {
    "seqname": "Chromosome",
    "source": "Source",
    "feature": "Feature",
    "start": "Start",
    "end": "End",
    "score": "Score",
    "strand": "Strand",
    "frame": "Frame",
    # "attribute": "Attribute"  # filled with all others columns
}

_gff3_columns = _gtf_columns.copy()
_gff3_columns["phase"] = _gff3_columns.pop("frame")

_ordered_gtf_columns = [
    "seqname",
    "source",
    "feature",
    "start",
    "end",
    "score",
    "strand",
    "frame",
    "attribute",
]
_ordered_gff3_columns = [
    "seqname",
    "source",
    "feature",
    "start",
    "end",
    "score",
    "strand",
    "phase",
    "attribute",
]


def _fill_missing(df, all_columns):
    columns = list(df.columns)

    if not df.get(all_columns) is None:
        outdf = df.get(all_columns)
    else:
        missing = set(all_columns) - set(columns)
        missing_idx = {all_columns.index(m): m for m in missing}
        not_missing = set(columns).intersection(set(all_columns))
        not_missing_ordered = sorted(not_missing, key=all_columns.index)
        outdf = df.get(not_missing_ordered)

        for idx, missing in sorted(missing_idx.items()):
            outdf.insert(idx, missing, ".")

    return outdf


def _bed(df, keep):
    all_columns = "Chromosome Start End Name Score Strand".split()

    outdf = _fill_missing(df, all_columns)

    noncanonical = set(df.columns) - set(all_columns)
    noncanonical = [c for c in df.columns if c in noncanonical]

    if keep:
        return pd.concat([outdf, df.get(noncanonical)], axis=1)
    else:
        return outdf


def _gtf(df, mapping):
    pr_col2gff_col = {v: k for k, v in mapping.items()}

    df = df.rename(columns=pr_col2gff_col)  # copying here
    df.loc[:, "start"] = df.start + 1
    all_columns = _ordered_gtf_columns[:-1]
    columns = list(df.columns)

    outdf = _fill_missing(df, all_columns)

    if "attribute" in df.columns:
        attribute = mapping["attribute"] + ' "' + df.attribute + '";'
    else:
        # gotten all needed columns, need to join the rest
        rest = set(df.columns) - set(all_columns)
        rest = sorted(rest, key=columns.index)
        rest_df = df.get(rest).copy()
        for c in rest_df:
            col = rest_df[c]
            isnull = col.isnull()
            col = col.astype(str).str.replace("nan", "")
            new_val = c + ' "' + col + '";'
            rest_df.loc[:, c] = rest_df[c].astype(str)
            rest_df.loc[~isnull, c] = new_val
            rest_df.loc[isnull, c] = ""

        attribute = rest_df.apply(lambda r: " ".join([v for v in r if v]), axis=1)
    outdf.insert(outdf.shape[1], "attribute", attribute)

    return outdf


def _to_gtf(self, path=None, compression="infer", map_cols=None):
    mapping = _gtf_columns.copy()
    if map_cols:
        mapping.update(map_cols)

    gr = self

    outdfs = [_gtf(v, mapping) for k, v in sorted(gr.dfs.items())]

    if path:
        mode = "w+"
        for outdf in outdfs:
            outdf.to_csv(
                path,
                index=False,
                header=False,
                compression=compression,
                mode=mode,
                sep="\t",
                quoting=csv.QUOTE_NONE,
            )
            mode = "a"
    else:
        return "".join([outdf.to_csv(index=False, header=False, sep="\t", quoting=csv.QUOTE_NONE) for outdf in outdfs])


def _to_csv(self, path=None, sep=",", header=True, compression="infer"):
    gr = self

    if path:
        mode = "w+"
        for _, outdf in natsorted(gr.dfs.items()):
            outdf.to_csv(
                path,
                index=False,
                compression=compression,
                header=header,
                mode=mode,
                sep=sep,
                quoting=csv.QUOTE_NONE,
            )
            mode = "a"
            header = False
    else:
        return "".join(
            [
                outdf.to_csv(index=False, header=header, sep=sep, quoting=csv.QUOTE_NONE)
                for _, outdf in sorted(gr.dfs.items())
            ]
        )


def _to_bed(self, path=None, sep="\t", keep=True, compression="infer"):
    gr = self

    outdfs = natsorted(gr.dfs.items())
    outdfs = [_bed(df, keep) for _, df in outdfs]

    if path:
        mode = "w+"
        for outdf in outdfs:
            outdf.to_csv(
                path,
                index=False,
                header=False,
                compression=compression,
                mode=mode,
                sep="\t",
                quoting=csv.QUOTE_NONE,
            )
            mode = "a"

    else:
        res = "".join([outdf.to_csv(index=False, header=False, sep="\t", quoting=csv.QUOTE_NONE) for outdf in outdfs])
        return res


def _to_bigwig(self, path, chromosome_sizes, rpm=True, divide=False, value_col=None, dryrun=False):
    try:
        import pyBigWig  # type: ignore
    except ModuleNotFoundError:
        print(
            "pybigwig must be installed to create bigwigs. Use `conda install -c bioconda pybigwig` or `pip install pybigwig` to install it."
        )
        import sys

        sys.exit(1)

    if not divide:
        gr = self.to_rle(rpm=rpm, strand=False, value_col=value_col).to_ranges()
    else:
        gr = self.to_rle(rpm=rpm, strand=False, value_col=value_col)
        divide_by = self.to_rle(rpm=rpm, strand=False)
        c = gr / divide_by
        new_pyrles = {}
        for k, v in c.items():
            v.values = np.log2(v.values)
            v.defragment()
            new_pyrles[k] = v

        gr = c.defragment().to_ranges()

    unique_chromosomes = gr.chromosomes

    subset = ["Chromosome", "Start", "End", "Score"]

    gr = gr[subset].unstrand()

    gr = gr.sort()

    if dryrun:
        return gr

    if not isinstance(chromosome_sizes, dict):
        size_df = chromosome_sizes.df
        chromosome_sizes = {k: v for k, v in zip(size_df.Chromosome, size_df.End)}

    header = [(c, int(chromosome_sizes[c])) for c in unique_chromosomes]

    bw = pyBigWig.open(path, "w")
    bw.addHeader(header)

    for chromosome, df in gr:
        chromosomes = df.Chromosome.tolist()
        starts = df.Start.tolist()
        ends = df.End.tolist()
        values = df.Score.tolist()

        bw.addEntries(chromosomes, starts, ends=ends, values=values)


def _to_gff3(self, path=None, compression="infer", map_cols=None):
    mapping = _gff3_columns.copy()
    if map_cols:
        mapping.update(map_cols)

    gr = self

    outdfs = [_gff3(v, mapping) for k, v in sorted(gr.dfs.items())]

    if path:
        mode = "w+"
        for outdf in outdfs:
            outdf.to_csv(
                path,
                index=False,
                header=False,
                compression=compression,
                mode=mode,
                sep="\t",
                quoting=csv.QUOTE_NONE,
            )
            mode = "a"
    else:
        return "".join(
            [outdf.to_csv(index=False, header=False, sep="\t", quoting=csv.QUOTE_MINIMAL) for outdf in outdfs]
        )


def _gff3(df, mapping):
    pr_col2gff_col = {v: k for k, v in mapping.items()}

    df = df.rename(columns=pr_col2gff_col)  # copying here
    df.loc[:, "start"] = df.start + 1
    all_columns = _ordered_gff3_columns[:-1]
    columns = list(df.columns)

    outdf = _fill_missing(df, all_columns)

    if "attribute" in mapping:
        attribute_name = mapping["attribute"]
        attribute_value = df.attribute.iloc[0]
        attribute = f"{attribute_name}={attribute_value}"
    else:
        # gotten all needed columns, need to join the rest
        rest = set(df.columns) - set(all_columns)
        rest = sorted(rest, key=columns.index)
        rest_df = df.get(rest).copy()
        total_cols = rest_df.shape[1]
        for i, c in enumerate(rest_df, 1):
            col = rest_df[c]
            isnull = col.isnull()
            col = col.astype(str).str.replace("nan", "")
            if i != total_cols:
                new_val = c + "=" + col + ";"
            else:
                new_val = c + "=" + col
            rest_df.loc[:, c] = rest_df[c].astype(str)
            rest_df.loc[~isnull, c] = new_val
            rest_df.loc[isnull, c] = ""

        attribute = rest_df.apply(lambda r: "".join([v for v in r if v]), axis=1).str.replace(";$", "", regex=True)
    outdf.insert(outdf.shape[1], "attribute", attribute)

    return outdf


# def _to_bam(df, filename, header=None, chromsizes=None):

#     def _header_from_chromsizes(chromsizes):

#         try:
#             chromsizes = {k: v for k, v in zip(chromsizes.Chromosome, chromsizes.End)}
#         except:
#             pass

#         chromosomes = []
#         for chromosome, length in chromsizes.items():
#             chromosomes.append({"LN": length, "SN": chromosome})

#         return {"SQ": chromosomes}

#     if chromsizes:
#         header = _header_from_chromsizes(chromsizes)


#     import sys

#     try:
#         import bamread
#     except ModuleNotFoundError as e:
#         print("bamread must be installed to write bam. Use `conda install -c bioconda bamread` or `pip install bamread` to install it.")
#         sys.exit(1)



import pyranges as pr


# 3' and 5'
def test_slack():
    gr = pr.PyRanges(chromosomes="chr1", starts=[15, 300], ends=[20, 305], strands="+ -".split())
    print(gr)
    gr = gr.slack({"5": 10, "3": 5})



import numpy as np  # NOQA: E402
import pandas as pd
from sorted_nearest import k_nearest_next_nonoverlapping, k_nearest_previous_nonoverlapping  # type: ignore


def nearest_previous_idx(d1, d2, k, ties=None):
    d1s = d1.Start.sort_values()
    d2e = d2.End.sort_values()

    # ix where d1s could be inserted into d2e to preserve order of d2e
    # start = time()
    ix = np.searchsorted(d2e, d1s, side="right") - 1
    # end = time()
    # print("np previous", end - start)

    valid = ix >= 0
    ix = ix[valid]

    d1s = d1s.iloc[valid]

    lidx, ridx_pos, dist = k_nearest_previous_nonoverlapping(d1s.values, d2e.values, d1s.index.values, ix, k, ties)

    dist += 1

    ridx = d2e.iloc[ridx_pos].index

    return lidx, ridx, dist


# @profile
def nearest_next_idx(d1, d2, k, ties=None):
    d1e = d1.End.sort_values()
    d2s = d2.Start.sort_values()

    # ix where d1e could be inserted into d2s to preserve order of d2s
    # start = time()
    ix = np.searchsorted(d2s, d1e, side="left")
    # end = time()
    # print("np next", end - start)

    # - print(ix)
    valid = ix < len(d2s)
    ix = ix[valid]

    d1e = d1e.iloc[valid]

    lidx, ridx_pos, dist = k_nearest_next_nonoverlapping(d1e.values, d2s.values, d1e.index.values, ix, k, ties)

    dist += 1

    ridx = d2s.iloc[ridx_pos].index

    return lidx, ridx, dist


def nearest(d1, d2, **kwargs):
    suffix = kwargs.get("suffix", "_b")
    ties = kwargs.get("ties", None)

    plidx, pridx, pdist = nearest_previous_idx(d1, d2, d1.__k__.values, ties)
    nlidx, nridx, ndist = nearest_next_idx(d1, d2, d1.__k__.values, ties)

    pk = d1.__k__.reindex(plidx)
    nk = d1.__k__.reindex(nlidx)
    # px = d1.__IX__.reindex(plidx)
    # nx = d1.__IX__.reindex(nlidx)

    p = pd.DataFrame({"LX": plidx, "RX": pridx, "D": pdist, "PN": "P", "k": pk})
    n = pd.DataFrame({"LX": nlidx, "RX": nridx, "D": ndist, "PN": "N", "k": nk})

    df = pd.concat([p, n]).sort_values(["LX", "D"])
    # print(df)
    # raise

    if df.empty:
        return None

    xdf = df

    # row_indexer = xdf.PN == "P"
    # xdf.loc[row_indexer, "D"] = -xdf[row_indexer].D

    d1 = d1.reindex(xdf.LX)
    d2 = d2.reindex(xdf.RX)
    d1.index = range(len(d1))
    d2.index = range(len(d1))
    d2 = d2.drop("Chromosome", axis=1)
    df = d1.join(d2, rsuffix=suffix)
    df.insert(df.shape[1], "Distance", xdf.D.values)

    # to_drop = [c for c in df.columns if "__k__" in c]

    # df = df.drop(to_drop, axis=1)

    return df


def nearest_previous(d1, d2, **kwargs):
    suffix = kwargs.get("suffix", "_b")
    ties = kwargs.get("ties", None)

    lidx, ridx, dist = nearest_previous_idx(d1, d2, d1.__k__.values, ties)

    d1 = d1.reindex(lidx)
    d2 = d2.reindex(ridx)
    d1.index = range(len(d1))
    d2.index = range(len(d1))

    d2 = d2.drop("Chromosome", axis=1)
    df = d1.join(d2, rsuffix=suffix)
    df.insert(df.shape[1], "Distance", dist)

    # df = remove_duplicates_single(df, ties)

    # df.loc[:, "Distance"] = - df.Distance
    # to_drop = [c for c in df.columns if "__k__" in c]
    # df = df.drop(to_drop, axis=1)

    return df


def nearest_next(d1, d2, **kwargs):
    suffix = kwargs.get("suffix", "_b")
    ties = kwargs.get("ties", None)

    lidx, ridx, dist = nearest_next_idx(d1, d2, d1.__k__.values, ties)

    d1 = d1.reindex(lidx)
    d2 = d2.reindex(ridx)
    d2 = d2.drop("Chromosome", axis=1)
    d1.index = range(len(d1))
    d2.index = range(len(d1))
    df = d1.join(d2, rsuffix=suffix)
    df.insert(df.shape[1], "Distance", dist)

    return df


# @profile
def _nearest(d1, d2, **kwargs):
    if d1.empty or d2.empty:
        return None

    how = kwargs["how"]

    # if kwargs.get("strandedness") == "opposite":
    #     strand_dict = {"+": "-", "-": "+"}
    # else:
    #     strand_dict = {"+": "+", "-": "-"}

    if how in ["upstream", "downstream"] and kwargs["stranded"]:
        strand = d1.Strand.iloc[0]
        # strand = strand_dict[strand]

        __nearest = {
            ("+", "upstream"): nearest_previous,
            ("-", "upstream"): nearest_next,
            ("+", "downstream"): nearest_next,
            ("-", "downstream"): nearest_previous,
        }[strand, how]
    elif how in ["upstream", "downstream"] and not kwargs["stranded"]:
        __nearest = {"upstream": nearest_previous, "downstream": nearest_next}[how]
    else:
        __nearest = nearest

    df = __nearest(d1, d2, **kwargs)

    return df



from pandas.testing import assert_series_equal

import pyranges as pr


# with slack
def test_join_with_slack():
    gr1 = pr.PyRanges(chromosomes="chr1", starts=[0], ends=[10], strands="+")
    gr2 = pr.PyRanges(chromosomes="chr1", starts=[15], ends=[20], strands="+")

    result = gr1.join(gr2, slack=10)
    df = result.df
    print(df)
    assert not df.empty


def test_join_without_reordering():
    f1 = pr.from_dict(
        {
            "Chromosome": ["chr1", "chr1", "chr1"],
            "Start": [3, 8, 5],
            "End": [6, 9, 7],
            "Name": ["interval1", "interval3", "interval2"],
        }
    )
    f2 = pr.from_dict({"Chromosome": ["chr1", "chr1"], "Start": [1, 6], "End": [2, 7], "Name": ["a", "b"]})

    lj = f1.join(f2, how="left", preserve_order=True)
    assert_series_equal(lj.Name, f1.Name)

    rj = f1.join(f2, how="right", preserve_order=True)
    assert_series_equal(rj.Name_b, f2.Name, check_names=False)

    oj = f1.join(f2, how="outer", preserve_order=True)
    assert list(oj.Name) == ["interval1", "interval3", "interval2", "-1"]
    assert list(oj.Name_b) == ["-1", "-1", "b", "a"]



import pandas as pd

import pyranges as pr

exons = pr.data.exons()
cpg = pr.data.cpg()
cpg.join(exons.unstrand()).subset(lambda df: df.CpG > 25)[["CpG"]].assign(lambda df: df.CpG % 10, "CpGDecile")[
    "chrX"
].slack(500)
from piedpiper import Debug

with Debug():
    cpg.join(exons.unstrand()).subset(lambda df: df.CpG > 25)[["CpG"]].assign(lambda df: df.CpG % 10, "CpGDecile")[
        "chrX"
    ].slack(500)



"""Tests that all example data loads properly."""

from pyranges.data import aorta, aorta2, chipseq, chipseq_background, cpg, ensembl_gtf, exons, f1, f2  # control_bam)


def test_all_data():
    for f in [
        f1,
        f2,
        chipseq,
        chipseq_background,
        cpg,
        exons,
        aorta,
        aorta2,
        ensembl_gtf,  # , control_bam
    ]:
        f()
